from collections import defaultdict
from shushuo.api import ShushuoApi
from shushuo.event import Event
from shushuo.profile import Profile
from shushuo.query import Query
from shushuo import exceptions


class ShuClient(object):

    def __init__(self, project_id, project_key, endpoint_url=None,
                 api_class=ShushuoApi, get_timeout=60, post_timeout=10):
        super(ShuClient, self).__init__()

        # Set up an api client to be used for querying and optionally passed
        # into a default persistence strategy.
        self.api = api_class(project_id, project_key,
                             endpoint_url=endpoint_url,
                             get_timeout=get_timeout,
                             post_timeout=post_timeout)

        self.project_id = project_id

    def push_events(self, collection, event_body):
        event = Event(collection, event_body)
        return self.api.post_event(event.collection_name,
                                   event.body)

    def mpush_events(self, events):
        if not isinstance(events, (dict)):
            raise exceptions.InvalidEventError(
                "Events must be a list of dict objects")

        events_by_collection = defaultdict(list)
        for collection in events:
            if isinstance(events[collection], list):
                for event in events[collection]:
                    e = Event(collection, event)
                    events_by_collection[collection].append(e.body)
            else:
                raise exceptions.InvalidEventError(
                    "Events must be a list of dict objects")

        return self.api.post_events(events_by_collection)

    def query_events(self, collection, query_body):
        query = Query(collection, query_body)

        return self.api.query_events(query.collection_name,
                                     query.body)

    query = query_events
    push = push_events
    mpush = mpush_events

    def update_profile(self, collection, profile_body):
        profile = Profile(collection, profile_body)
        return self.api.post_profile(profile.collection_name,
                                     profile.body)



import re
import copy
from datetime import datetime
from shushuo import exceptions


Reserved_fields = [
    '@id',
    '@timestamp',
    '@location',
    '@actor',
]


class Event(object):
    """
    Event is the object used to validate and process event data before
    sending it to the Shushuo API
    """

    def __init__(self, collection_name, event_data):
        if not re.match('^[a-zA-Z_][a-zA-Z0-9_]*$', collection_name):
            raise exceptions.InvalidEventError(
                'Collection Name is not valid.')

        self.collection_name = collection_name

        self.body = self._process(event_data)

    def _process(self, event):
        """
        Checks if an event is valid for Shushuo and returns processed event
        that is safe to send to api.
        Raises InvalidEventError if the event is invalid

        :param event_data: dictionary with data for the event
        """
        if not isinstance(event, dict):
            raise exceptions.InvalidEventError('Event must be a dict object.')

        event_copy = copy.deepcopy(event)

        if '@timestamp' not in event_copy:
            event_copy['@timestamp'] = datetime.utcnow()

        if not isinstance(event_copy['@timestamp'], datetime):
            raise exceptions.InvalidEventError(
                'Timestamp must be a datetime object.')

        if '@id' in event_copy:
            raise exceptions.InvalidEventError('Event should not have id.')

        self._validate(event_copy)
        return event_copy

    def _validate(self, event, nested_depth=0):
        depth = nested_depth
        for key, value in event.items():
            if not isinstance(key, (str, unicode)):
                raise exceptions.InvalidEventError(
                    'Event key not valid %s' % key)
            elif nested_depth == 0 and key in Reserved_fields:
                pass
            elif not re.match('^[a-zA-Z][a-zA-Z0-9_]*$', key):
                raise exceptions.InvalidEventError(
                    'Event key not valid %s' % key.encode('utf-8'))

            if value is None:
                continue

            if isinstance(value, (unicode, str, long,
                                  int, float, bool)):
                continue
            elif isinstance(value, datetime):
                event[key] = value.isoformat()
            elif isinstance(value, dict):
                if key == '_location' and nested_depth == 0:
                    if len(value) == 2 and \
                       value.get('lat', None) and \
                       value.get('lon', None):
                        pass
                    else:
                        raise exceptions.InvalidEventError(
                            'invalid location format %s' % value)
                else:
                    _depth = self._validate(value,
                                            nested_depth=nested_depth+1)
                    depth = max(depth, _depth)
            elif isinstance(value, list):
                for item in value:
                    if not isinstance(item, (str, unicode)):
                        raise exceptions.InvalidEventError(
                            'invalid location format %s' % value)
            else:
                raise exceptions.InvalidEventError(
                    'unkown value type %s' % type(value))
        return depth



import json
from requests import Session
from shushuo import response


class ShushuoApi(object):
    """
    Used by the client to communicate to the shushuo API.
    """

    def __init__(self, project_id, project_key,
                 endpoint_url=None,
                 post_timeout=None,
                 get_timeout=None):
        self._project_id = project_id
        self._project_key = project_key
        self._session = self._create_session()

        if post_timeout is not None:
            self._post_timeout = post_timeout
        else:
            self._post_timeout = 60

        if get_timeout is not None:
            self._get_timeout = get_timeout
        else:
            self._get_timeout = 60

        if endpoint_url is not None:
            self._endpoint_url = endpoint_url
        else:
            self._endpoint_url = "https://api.shushuo.com"

    def _create_session(self):
        s = Session()
        s.headers.update({
            "Content-Type": "application/json",
            "X-Project-Key": self._project_key,
            "X-Project-Id": self._project_id
        })
        return s

    def post_event(self, collection_name, event):
        url = "{0}/events/{1}".format(self._endpoint_url, collection_name)
        payload = json.dumps(event)

        r = self._session.post(url=url,
                               data=payload,
                               timeout=self._post_timeout)

        return self._build_response(r)

    def post_events(self, events):
        url = "{0}/events".format(self._endpoint_url)
        payload = json.dumps(events)

        r = self._session.post(url=url,
                               data=payload,
                               timeout=self._post_timeout)

        return self._build_response(r)

    def post_profile(self, collection_name, profile):
        url = "{0}/profiles/{1}".format(self._endpoint_url, collection_name)
        payload = json.dumps(profile)

        r = self._session.post(url=url,
                               data=payload,
                               timeout=self._post_timeout)

        return self._build_response(r)

    def query_events(self, collection_name, query_body):
        url = "{0}/events/{1}".format(self._endpoint_url, collection_name)
        payload = {
            'query': json.dumps(query_body)
        }

        r = self._session.get(url=url,
                              params=payload,
                              timeout=self._get_timeout)

        return self._build_response(r)

    def _build_response(self, r):
        try:
            result = r.json()
        except ValueError:
            result = {
                'error_code': r.status_code,
                'message': r.text
            }

        error_message = None
        if r.status_code == 403:
            error_message = "Unauthorised. Plz check your Project Id and Key"
        elif r.status_code != 200:
            error_message = result.get('message', 'Unknow issue.')

        return response.ShushuoResponse(status_code=r.status_code,
                                        error_message=error_message,
                                        result=result)



import re
import copy
from shushuo import exceptions

Reserved_aggregations = [
    'max',
    'min',
    'avg',
    'sum',
    'count',
    'unique',
    'last',
]

Reserved_timeframes = [
    'this_minute',
    'last_minute',
    'this_hour',
    'last_hour',
    'today',
    'yesterday',
    'this_week',
    'last_week',
    'this_month',
    'last_month',
    'this_year',
    'last_year'
]

Reserved_intervals = [
    'minute',
    'hour',
    'day',
    'week',
    'month',
    'year',
]

Reserved_fields = [
    'select',
    'filter',
    'interval',
    'group_by',
    'timeframe',
    'timezone',
]


class Query(object):
    """
    Query is the object used to validate and process event data before
    sending it to the Shushuo API
    """

    def __init__(self, collection_name, query_data):
        if not re.match('^[a-zA-Z_][a-zA-Z0-9_]*$', collection_name):
            raise exceptions.InvalidEventError(
                'Collection Name is not valid.')

        self.collection_name = collection_name
        self.body = self._process(query_data)

    def _process(self, query):
        if not isinstance(query, dict):
            raise exceptions.InvalidQueryError('Query must be a dict object.')

        unknow_keys = set(query.keys()) - set(Reserved_fields)
        if unknow_keys:
            raise exceptions.InvalidQueryError(
                'Query should not contain %s.' % unknow_keys)

        if 'select' not in query:
            raise exceptions.InvalidQueryError(
                'Query should atleast contain select.')

        query_copy = copy.deepcopy(query)

        for field in query_copy:
            query_copy[field] = getattr(self, '_process_%s' % field)(
                query_copy.get(field, None)
            )

        return query_copy

    def _process_select(self, select):
        if select == '*':
            select = ['*']

        if select == 'count(*)':
            select = {
                'count': 'count'
            }

        if isinstance(select, dict):
            for label in select:
                aggregate = select[label]
                if aggregate == 'count':
                    continue

                if isinstance(aggregate, dict) and len(aggregate.keys()) == 1:
                    aggregation, field = aggregate.items()[0]
                    if aggregation not in Reserved_aggregations:
                        raise exceptions.InvalidQueryError(
                            'Unknown aggregation %s.' % aggregation)
                    if not isinstance(field, (str, unicode)):
                        raise exceptions.InvalidQueryError(
                            'Query field %s should be string.' % field)
                else:
                    raise exceptions.InvalidQueryError(
                        'Invalid Aggregate format %s.' % aggregate)
        elif isinstance(select, list):
            if len(select) == 0:
                raise exceptions.InvalidQueryError(
                    'Select should not be empty.')
            for field in select:
                if not isinstance(field, (str, unicode)):
                    raise exceptions.InvalidQueryError(
                        'Query field %s should be string.' % field)
        else:
            raise exceptions.InvalidQueryError(
                'Select format should be list or dictionary.')

        return select

    def _process_timezone(self, timezone):
        if timezone is None:
            return timezone

        if not isinstance(timezone, int):
            raise exceptions.InvalidQueryError(
                'Timezone field %s should be int.' % timezone)
        return timezone

    def _process_timeframe(self, timeframe):
        if timeframe is None:
            return timeframe

        if isinstance(timeframe, str):
            if timeframe not in Reserved_timeframes:
                raise exceptions.InvalidQueryError(
                    'Unkonw timeframe value %s.' % timeframe)
        elif isinstance(timeframe, dict):
            pass
        else:
            raise exceptions.InvalidQueryError(
                'Unkonw Timeframe format %s.' % timeframe)

        return timeframe

    def _process_interval(self, interval):
        if interval is None:
            return interval

        if interval not in Reserved_intervals:
            raise exceptions.InvalidQueryError(
                'Unkonw Interval value %s.' % interval)
        return interval

    def _process_filter(self, _filter):
        return _filter

    def _process_group_by(self, group_by):
        if group_by is None:
            return group_by

        if not isinstance(group_by, str):
            raise exceptions.InvalidQueryError(
                'GroupBy field %s should be string.' % group_by)
        return group_by





class BaseError(Exception):

    def __str__(self):
        # all sub-classes should set self._message in their initializers
        return self._message


class InvalidEventError(BaseError):

    def __init__(self, message):
        super(InvalidEventError, self).__init__()

        self.message = message
        self._message = "Invalid event: {0}".format(message)


class InvalidProfileError(BaseError):

    def __init__(self, message):
        super(InvalidProfileError, self).__init__()

        self.message = message
        self._message = "Invalid profile: {0}".format(message)


class InvalidQueryError(BaseError):

    def __init__(self, message):
        super(InvalidQueryError, self).__init__()

        self.message = message
        self._message = "Invalid query: {0}".format(message)





class ShushuoResponse(object):

    def __init__(self, status_code, error_message, result):
        self.success = status_code == 200
        self.status_code = status_code
        self.error_message = error_message
        self.result = result



#!/usr/bin/env python

from setuptools import setup
import os
import sys

setup_path = os.path.dirname(__file__)
tests_require = ['nose', 'mock']

if sys.version_info < (2, 7):
    tests_require.append('unittest2')

setup(
    name="shushuo",
    version="0.0.8",
    description="Python Client for ShuShuo.com",
    author="Kerwin",
    author_email="piaoyuankui@gmail.com",
    url="https://github.com/shushuo/shushuo.py",
    packages=["shushuo"],
    install_requires=[
        "requests",
        "urllib3"
    ],
    tests_require=tests_require,
)
